"""Core entities: base classes and full implementations."""
# TODO: Some of the docstrings parts should be moved to docs.
# TODO: Add logging.
# TODO: Fix snippet formatting in docstrings.

from typing import TypeVar, Dict, Optional, Union

from pydantic import BaseSettings

from nodzz.basic_types import JSONType, JSONDict, NodeStatus


# TODO: Implement module/name fields based configuration.
# TODO: Implement caption field.
class ConfigModelBase(BaseSettings):
    """Base model for component configs.

    Any ``nodzz`` component config can be JSON serializable dict (``JSONDict``) or
    ``pydantic`` BaseSettings model. If second, config model should be derived
    from this class. Fields defined on the ``nodzz`` library level should not be
    overridden or used for the logic they were not designed for.

    ``JSONDict`` representation allows to manage ``nodzz`` configuration via JSON
    and YAML files, database or even graphic tools.

    ``ConfigModelBase`` based component config models offer such features as default
    values definition, better autodocumentation and autocompletion in IDE as
    well as ``pydantic`` features like validation and autofilling parameters
    values from environment variables. Component config model can be defined
    by assigning config model class as a value to the ``config_model`` component
    class-level variable. We are strongly encouraging you to define config model
    for each component you create in your ``nodzz`` powered projects.

    Attributes:
        name: Str with component name.
        class_name: Str with full import path to the python class which implements
            component. Field is used only when parsing ``JSONDict`` config. Mutually
            exclusive with ``component_name`` attribute.
        component_name: is a reference to the name of some other ``nodzz`` component
            in the configuration namespace. Field is used only when parsing ``JSONDict``
            config.  Mutually exclusive with ``class_name`` attribute.
    """
    name: Optional[str] = None
    class_name: Optional[str] = None
    component_name: Optional[str] = None


TConfigModel = TypeVar('TConfigModel', bound=ConfigModelBase)
TConfig = Union[JSONDict, TConfigModel]


class ComponentBase:
    """Base class for any ``nodzz`` project component.

    Component is a ``nodzz`` project atomic building block (like node or some connector,
    for example). Any ``nodzz`` component should derive from this class. If component is
    configurable, its config can be defined through the class derived from ``ConfigModelBase``
    (which is ``pydantic`` config models in fact or JSON serializable dict (``JSONDict``)
    config can be used.

    Attributes:
        config_model: Component config class derived from ``ConfigModelBase``.
            Can be used for default values storage. If config model is defined
            it will be initialised  from ``JSONDict`` configs if such is provided
            instead of config model instance.
        config: ``pydantic`` based or ``JSONDict`` component config.
    """
    config_model: Optional[TConfigModel] = None
    config: Optional[TConfig]

    # TODO: Think of more strict config validation.
    def __init__(self, config: Optional[TConfig] = None) -> None:
        """Initialises component config.

        Validates config model type and initialised component config:

        Args:
            config: ``pydantic`` based or ``JSONDict`` component config.
        """
        if not self.config_model:
            self.config = config
        else:
            if not issubclass(self.config_model, ConfigModelBase):
                raise TypeError('Config model should be a subclass of ComponentConfigBase')

            if isinstance(config, dict):
                self.config = self.config_model.parse_obj(config)
            elif config is None:
                self.config = self.config_model()
            else:
                self.config = config


TComponent = TypeVar('TComponent', bound=ComponentBase)


class State:
    """Represents state of any agent executing its behavior program.

    State contains both current behavior tree nodes statuses and arbitrary
    named variables values which can be read or modified either by any
    decision tree node or by means of communication of behavior agent
    with the external environment.

    Metaphorically, state represent snapshot of behavior agent consciousness
    in any given moment: it contains both inputs from the interaction with the
    external environment and results of these inputs processing.

    State variables are stored in the Dict which can be accessed through the
    ``vars`` property. Keys are variables names and values are variables values.
    There is an important project-wide convention: ``None`` state variable value
    always represents uninitialised variable.

    Also JSON is used as a default serialisation format in ``nodzz``, so all
    variable values are presumed to be JSON serializable. No actual type
    checking is implemented, though some default methods like conversion
    to/from dict are implemented in JSON friendly way. You can derive
    your own state from this class and implement it in the way appropriate
    to your project.
    """
    def __init__(self, uid: Optional[str] = None,
                 variables: Optional[Dict[str, JSONType]] = None,
                 nodes: Optional[Dict[str, NodeStatus]] = None) -> None:
        """Initialises state instance.

        State constructor is presumed to be used for blank state instance
        creation, so mostly you will not have to specify ``variables`` and
        ``nodes`` arguments. ``from_dict`` class method can be referred for
        these arguments usage.

        Args:
            uid: An optional str state unique id.
            variables: An optional dict with state variables value.
            nodes: An optional dict with state nodes statuses.
        """
        self._uid = uid
        self._vars = variables or {}
        self._nodes = nodes or {}

    @property
    def uid(self) -> Optional[str]:
        """An optional string unique id."""
        return self._uid

    @property
    def vars(self) -> JSONDict:
        """State variables dictionary."""
        return self._vars

    def to_dict(self, nodes: bool = False) -> JSONDict:
        """Converts state to its dict representation.

        Args:
            nodes: A boolean flag which includes nodes statuses into results.
                Since each node status is converted to the serialization friendly
                format, it can be some kind of expensive operation, so the default
                value of the flag is ``False``.

        Returns: Dict representation of the state.
        """
        result = {
            'uid': self._uid,
            'variables': self._vars
        }

        if nodes:
            result['nodes'] = {
                node: status.value for node, status in self._nodes.items() if status is not NodeStatus.READY
            }

        return result

    @classmethod
    def from_dict(cls, state_dict: JSONDict) -> 'State':
        """Initialises state from its dict representation.

        Args:
            state_dict: Dict representation of state.

        Returns:
            Instance of the State object.
        """
        uid = state_dict['uid']
        variables = state_dict['variables']
        nodes = {node: NodeStatus(status) for node, status in state_dict.get('nodes', {}).items()}
        return cls(uid=uid, variables=variables, nodes=nodes)

    def get_node_status(self, node_id: str) -> NodeStatus:
        """Returns status of a single node.

        Args:
            node_id: Str node id in behavior tree.

        Returns:
            Node status.
        """
        return self._nodes.get(node_id, NodeStatus.READY)

    def reset_node_status(self, node_id: str) -> None:
        """Resets single node status.

        Removes status of the single node from statuses dict which is equal
        to setting READY status to the node. Status removal (instead of READY
        assigning) allows to reduce amount of operations during state serialisation.

        Args:
            node_id: Str node id in behavior tree.
        """
        self._nodes.pop(node_id, None)

    def set_node_status(self, node_id: str, status: NodeStatus) -> None:
        """Sets single node status.

        Status removal (instead of READY assigning) allows to reduce amount
        of operations during state serialisation.

        Args:
            node_id: Str node id in behavior tree.
            status: Status assigned to the node.
        """
        if status is NodeStatus.READY:
            self.reset_node_status(node_id)
        else:
            self._nodes[node_id] = status

    def reset_nodes(self) -> None:
        """Resets statuses of ALL nodes for the state."""
        self._nodes = {}



"""Complex type annotations and basic data structures used throughout the project."""

from __future__ import annotations

from enum import Enum
from typing import Union, List, Dict


# JSON-serializable data types annotations.
JSONSimpleType = Union[int, float, bool, None, str]
JSONType = Union[JSONSimpleType, List['JSONType'], Dict[str, 'JSONType']]
JSONList = List[JSONType]
JSONDict = Dict[str, JSONType]
JSON = Union[JSONList, JSONDict]


class NodeStatus(Enum):
    READY = 0
    RUNNING = 1
    SUCCESS = 2
    FAILED = 3



"""Controller nodes implementations."""
# TODO: Some of the docstrings parts should be moved to docs.

from nodzz.core import State
from nodzz.basic_types import NodeStatus
from nodzz.nodes.base import ControllerNodeBase, TNode


class SelectorNode(ControllerNodeBase):
    """Selector node finds (selects) its first child node that is executed without fail.

    Node sequentially executes its child nodes until one of them returns SUCCESS
    or RUNNING status. This implementation is NON-PERSISTENT, so when one of the
    nodes returns RUNNING status, child nodes statuses are not saved and RUNNING
    status is immediately returned. The next node execution starts with the first
    child node.

    When one of the child nodes returns SUCCESS status, the same status is immediately
    returned without further children nodes execution. When all of the child nodes
    return FAILED status, the same status is immediately returned.

    If child nodes priority among each other matters, it can be set by defining
    theirs execution order during selector node initialisation.

    Selector node behavior resembles OR logical gate and usually used to select
    one of possible behaviors.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    def execute(self, state: State) -> NodeStatus:
        """Executes its child nodes until one of them returns RUNNING or SUCCESS status.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = NodeStatus.FAILED

        for node in self.children:
            status = node.execute(state)

            if status != NodeStatus.FAILED:
                break

        return status


class PersistentSelectorNode(ControllerNodeBase):
    """Selector node finds (selects) its first child node that is executed without fail.

    Node sequentially executes its child nodes until one of them returns SUCCESS
    or RUNNING status. This implementation is PERSISTENT, so when one of the
    nodes returns RUNNING status, child nodes statuses are saved and RUNNING
    status is immediately returned. The next node execution starts with the
    running child node.

    When one of the child nodes returns SUCCESS status, the same status is immediately
    returned without further children nodes execution. When all of the child nodes
    return FAILED status, the same status is immediately returned.

    If child nodes priority among each other matters, it can be set by defining
    theirs execution order during selector node initialisation.

    Selector node behavior resembles OR logical gate and usually used to select
    one of possible behaviors.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    def execute(self, state: State) -> NodeStatus:
        """Executes its child nodes until one of them returns RUNNING or SUCCESS status.

        If child node returns RUNNING status all child nodes statuses are saved until
        next sequence node execution.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = NodeStatus.FAILED

        for node in self.children:
            current_status = state.get_node_status(node_id=node.id)

            if current_status == NodeStatus.READY or current_status == NodeStatus.RUNNING:
                status = node.execute(state)

            if status == NodeStatus.FAILED or status == NodeStatus.RUNNING:
                state.set_node_status(node_id=node.id, status=status)

            if status == NodeStatus.RUNNING or status == NodeStatus.SUCCESS:
                break

        if status == NodeStatus.SUCCESS or status == NodeStatus.FAILED:
            self.reset(state)

        return status


class SequenceNode(ControllerNodeBase):
    """Sequence node executes its child nodes until one of them is executed without success.

    Sequence Node sequentially executes its child nodes until one of them returns
    FAILED or RUNNING status. This implementation is NON-PERSISTENT, so when one
    of the nodes returns RUNNING status, child nodes statuses are not saved and
    RUNNING status is immediately returned without further children nodes execution.
    The next node execution starts with the first child node.

    When one of the child nodes returns FAILED status, the same status is immediately
    returned without further children nodes execution. When all of the child nodes
    return SUCCESS status, the same status is immediately returned.

    Children nodes execution order is defined during sequence node initialisation.

    Sequence node behavior resembles AND logical gate and usually used to encapsulate
    sequence of actions needed to perform some task.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    def execute(self, state: State) -> NodeStatus:
        """Executes its child nodes until one of them returns RUNNING or FAILED status.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = NodeStatus.SUCCESS

        for node in self.children:
            status = node.execute(state)

            if status != NodeStatus.SUCCESS:
                break

        return status


class PersistentSequenceNode(ControllerNodeBase):
    """Sequence node executes its child nodes until one of them is executed without success.

    Sequence Node sequentially executes its child nodes until one of them returns
    FAILED or RUNNING status. This implementation is PERSISTENT, so when one of the
    nodes returns RUNNING status, ALL child nodes statuses are saved and RUNNING
    status is immediately returned without further children nodes execution. The
    next node execution starts with the running child node.

    When one of the child nodes returns FAILED status, the same status is immediately
    returned without further children nodes execution. When all of the child nodes
    return SUCCESS status, the same status is immediately returned. In both of these
    cases all child nodes statuses will be reset.

    Children nodes execution order is defined during sequence node initialisation.

    Sequence node behavior resembles AND logical gate and usually used to encapsulate
    sequence of actions needed to perform some task.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    def execute(self, state: State) -> NodeStatus:
        """Executes its child nodes until one of them returns RUNNING or FAILED status.

        If child node returns RUNNING status all child nodes statuses are saved until
        next sequence node execution.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = NodeStatus.SUCCESS

        for node in self.children:
            current_status = state.get_node_status(node_id=node.id)

            if current_status == NodeStatus.READY or current_status == NodeStatus.RUNNING:
                status = node.execute(state)

            if status == NodeStatus.SUCCESS or status == NodeStatus.RUNNING:
                state.set_node_status(node_id=node.id, status=status)

            if status == NodeStatus.RUNNING or status == NodeStatus.FAILED:
                break

        if status == NodeStatus.SUCCESS or status == NodeStatus.FAILED:
            self.reset(state)

        return status



"""Configuration management tools.

While ``nodzz`` have the instruments to implement, manage and configure any
behavior tree via pure Python API, this is of course not the most convenient
way to manage behavior trees based projects. ``nodzz`` provides config based
behavior tree management tools which allow completely separate tree graph
design and nodes configuring from each particular node Python implementation.
Here are the core concepts of ``nodzz`` configuration:

Component.

Component is a ``nodzz`` project atomic building block (like node
or some connector, for example).

In the ``nodzz`` configuration tools component is represented by unique name
which identifies it among all other components and its config. From here on,
the ter "component name" and the term "config name" will mean the same.
Component config is a JSON-serializable mapping data structure which can
exist both in the form of ``JSONDict`` (JSON-serializable dict) and in the
form of ``pydantic`` based ``ConfigModelBase`` config which in turn can be
one-to-one converted from or to the ``JSONDict`` representation. Please refer
``ConfigModelBase`` class documentation for more info.

Each component config must have one (and only one) of two fields initialised:
``class_name`` or ``component_name``.
* ``class_name`` is a full import path to the python class which implements
the component and which is initialized with the config. Import path should
should locate component class in current Python environment and has format
like ``nodzz.nodes.controllers.SequenceNode``. If ``class_name`` is initialised,
component config should provide all necessary parameters for the component
correct work.
* ``component_name`` is a reference to the name of some other ``nodzz`` component
(referenced component) in the configuration namespace. In this case referenced
config will be merged to the initial config: all uninitialized fields in the
initial component config will be automatically filled by corresponding values
from the referenced component config. This process is called "config resolution".
Also the component being referenced can itself have a reference to another
component. In such case config resolution process covers all configs chain.
This feature implements some kind of inheritance among the components configs
and encourages components reuse.

Config set.

Config set as a logical entity is a set (surprise!) of component configs that
exist in the same namespace. This allows to identify and access all component
configs by their names. Config set implemented in the config set entity.

"""
# TODO: Some of the docstrings parts should be moved to docs.
# TODO: Implement config dirs loading.

from pathlib import Path
from typing import Optional, List, Dict, Any

from pydantic import BaseModel

from nodzz.basic_types import JSONDict
from nodzz.utils import load_file


class ConfigSet:
    """Component configs management entity.

    ``ConfigSet`` is a container for component configs, loaded from external sources
    like files or databases. It provides instruments for configs access, management
    and validation. ``ConfigSet`` serves as a library of component configs. All configs
    required for any behavior initialisation tree should be contained in one instance
    of ``ConfigSet``.
    """
    def __init__(self) -> None:
        self._meta_configs = {}

    def add_config(self, name: str, config: JSONDict, source: str = 'undefined', update: bool = False) -> None:
        """Adds component config to the config set.

        Adds component config and its metadata to the config set, validates and
        optionally resolves config.

        Args:
            name: Config name.
            config: Component config in JSON-serializable dict format.
            source: Config source ID, used for config source identification in case
                of config management problems.
            update: Boolean flag controlling the behavior when config with already
                existing name is being added. If value is ``True``, existing config will
                be replaced by the new one. If value is ``False``, exception will be raised.
                Default value is ``False``.
        """
        class_name = config.get('class_name')
        ref_name = config.get('component_name')

        if class_name and ref_name:
            raise ValueError(f'Both class_name and component_name are defined in config: {name}(source: {source})')
        elif not class_name and not ref_name:
            raise ValueError(f'Both class_name and component_name are not defined in config: {name} (source: {source})')

        if name in self._meta_configs and not update:
            raise ValueError(f'Trying to add config with already existing name: {name}, '
                             f'new config source: {source}, '
                             f'existing config source: {self._meta_configs[name]["source"]}')

        # TODO: Now is assigning used only for debug mode. Remove when name/module configuration will be implemented.
        config['name'] = name

        self._meta_configs[name] = {
            'config': config,
            'source': source
        }

    def get_config(self, name: str) -> Optional[JSONDict]:
        """Returns config from the config set by its name.

        Args:
            name: Config name.

        Returns:
            JSONDict component config if config with given name exists in
            config set, else ``None``.
        """
        meta_config = self._meta_configs.get(name)
        result = meta_config['config'] if meta_config else None
        return result

    def del_config(self, name: str) -> None:
        """Deletes config from the config set by its name.

        Args:
            name: Config name.
        """
        self._meta_configs.pop(name, None)

    def _resolve_config(self, name: str, config: JSONDict, source: str, chain: Optional[List[str]] = None) -> JSONDict:
        """Resolves component config.

        Recursively resolves all configs in reference chain of given config (including
        given config itself). References are looked among already loaded to the config
        set configs.

        Args:
            name: Config name.
            config: Component config in JSON-serializable dict format.
            source Config source ID, used for config source identification in case
                of config management problems.
            chain: List which accumulates non-resolved config names in the method
                recursive calls cycle. Used for the cyclic references detection in
                the component configs.

        Returns:
            Dict containing all resolved configs from the reference chain of given
            config. Keys are config names and values are corresponding configs. Empty
            dict is returned when no config was resolved.
        """
        result = {}
        ref_name = config.get('component_name')

        if ref_name:
            ref_meta_config = self._meta_configs.get(ref_name)

            if ref_meta_config is None:
                raise ValueError(f'Component {name} (source: {source}) refers to the absent component {ref_name}')

            chain = chain or []

            if ref_name in chain:
                error_message = 'Cyclic reference among component configs: '

                for _name in chain:
                    _ref_name = self._meta_configs[_name]['config']['component_name']
                    _source = self._meta_configs[_name]['source']
                    info_str = f'\n\tname: {_name}, refers to: {_ref_name}, source: {_source}'
                    info_str = f'\n{info_str}' if _name == ref_name else info_str
                    error_message = f'{error_message}{info_str}'

                info_str = f'\n\tname: {name}, refers to: {ref_name}, source: {source}'
                error_message = f'{error_message}{info_str}'

                raise ValueError(f'Cyclic reference among components:\n{error_message}')

            ref_config = ref_meta_config['config']
            ref_ref_name = ref_config.get('component_name')

            if ref_ref_name:
                chain.append(name)
                ref_source = ref_meta_config['source']
                result = self._resolve_config(ref_name, ref_config, ref_source, chain)
                ref_config = result[ref_name]

            resolved_config = {k: v for k, v in ref_config.items()}
            resolved_config.update(**config)
            resolved_config.pop('component_name')

            result[name] = resolved_config

        return result

    def resolve_configs(self) -> None:
        """Prepares all config set configs for components initialisation.

        "Flattens" config set structure: resolves all references between
        configs by setting field values which are absent in config with
        corresponding field values from referenced config.
        """
        for name, meta_config in self._meta_configs.items():
            config = meta_config['config']
            source = meta_config['source']
            resolved_configs = self._resolve_config(name=name, config=config, source=source)

            for config_name, resolved_config in resolved_configs.items():
                self._meta_configs[config_name]['config'] = resolved_config


class _ConfigFileModel(BaseModel):
    """Config file validation model."""
    __root__: Dict[str, Dict[str, Any]]


def load_config_file(file_path: Path) -> ConfigSet:
    """Loads component configs from single JSON/YAML file and initialises ``ConfigSet`` entity.

    File should contain dict as a root element where keys are config names and
    values are configs.

    Args:
        file_path: ``pathlib`` object with configuration file path.

    Returns:
        ``ConfigSet`` instance with resolved configs loaded from file.
    """
    config_dict = load_file(file_path)
    _ConfigFileModel.parse_obj(config_dict)
    config_set = ConfigSet()
    file_path_str = str(file_path)

    for name, config in config_dict.items():
        config_set.add_config(name=name, config=config, source=file_path_str)

    config_set.resolve_configs()

    return config_set



"""Typical task nodes implementations."""
# TODO: Some of the docstrings parts should be moved to docs.

from collections import defaultdict
from functools import partial
from typing import List, Union, Optional, TypeVar, Dict

from pydantic import BaseModel, validator, Field

from nodzz.basic_types import JSONSimpleType, JSONType, NodeStatus
from nodzz.core import ConfigModelBase, State
from nodzz.nodes.base import NodeBase

_EvalArgs = Union[List[JSONSimpleType], float, int, bool, str]
_EvalArgsNone = Union[_EvalArgs, None]


class EvalSettingsBase(BaseModel):
    """Base model for evaluation settings.

    Attributes:
        eval_type: Str evaluation type identifier, its constant default value must be
            set in subclasses.
        invert: Boolean inversion flag. If ``True``, ``not`` operator is applied to the
            evaluation result
    """
    eval_type: str
    invert: bool = False

    @validator('eval_type')
    def match_eval_type(cls, v: str) -> str:
        """Validates evaluation type name.

        Method is used for validation during config initialisation from
        JSON-like dicts by ``parse_obj`` method. Ensures proper evaluation
        settings initialisation.

        Args:
            v: Str ``eval_type`` field value.

        Returns:
            Unchanged str ``eval_type`` field value.
        """
        eval_type = cls.__fields__['eval_type'].default

        if v != eval_type:
            raise ValueError(f'Not valid evaluation type \'{v}\', should be \'{eval_type}\'')

        return v


class Equal(EvalSettingsBase):
    """'Is equal' evaluation config model.

    Configures evaluation: if variable value is equal to the given value.

    Attributes:
        value: Value to be compared with variable value.
    """
    eval_type: str = Field(default='equal', const=True)
    value: _EvalArgs


class More(EvalSettingsBase):
    """'Is more' evaluation config model.

    Configures evaluation: if variable value is more then the given value.

    Attributes:
        value: Value to be compared with variable value.
        strict: Boolean comparison type. If ``True`` - strict comparison will be
            applied, if ``False`` - not strict.
    """
    eval_type: str = Field(default='more', const=True)
    value: Union[int, float]
    strict: bool = True


class Less(EvalSettingsBase):
    """'Is less' evaluation config model.

    Configures evaluation: if variable value is less then the given value.

    Attributes:
        value: Value to be compared with variable value.
        strict: Boolean comparison type. If ``True`` - strict comparison will be
            applied, if ``False`` - not strict.
    """
    eval_type: str = Field(default='less', const=True)
    value: Union[int, float]
    strict: bool = True


class Intersection(EvalSettingsBase):
    """'Lists intersection' evaluation config model.

    Configures evaluation: if list from variable value has at least one element
    in common with the given list.

    Attributes:
        value: Value to be compared with variable value.
    """
    eval_type: str = Field(default='intersection', const=True)
    value: List[JSONSimpleType]


_SingleEvaluation = Union[Equal, More, Less, Intersection]
_Evaluation = Union[_SingleEvaluation, List[_SingleEvaluation]]


class EvaluationNodeConfig(ConfigModelBase):
    """Base model for the ``EvaluationNode`` settings.

    Every ``EvaluationNode`` config should be derived from this class. Each
    variable value evaluation is configured by an item of the ``conditions``
    field dict, where string key is equal to the evaluated variable name
    and value has two possible type options:
    1. One of ``EvalSettingsBase`` subclasses (but not the ``EvalSettingsBase``
        itself). In this case only one variable evaluation will be performed
        according to the evaluation type and values it was initialised with.
    2. List of ``EvalSettingsBase`` subclasses. In this case each defined
        evaluation will be performed for the variable.

    Attributes:
        conditions: Dict with evaluation conditions.
        eval_none: Bool uninitialised values evaluation flag. If ``True``
            uninitialised variables (which are equal to variables with
            ``None`` values) will be estimated. If ``False``, node will return
            RUNNING status until all variables will be initialised or
            at least one evaluation fails.
    """
    conditions: Dict[str, _Evaluation]
    eval_none: bool = False


TEvaluationNodeConfig = TypeVar('TEvaluationNodeConfig', bound=EvaluationNodeConfig)


class EvaluationNode(NodeBase):
    """Node that evaluates state variables values.

    This node implements one of the most common tasks: check if some state
    variables values meet certain conditions. ``EvaluationNode`` allows to
    parametrise all variables values evaluations in node config, so its
    enough to create new config by deriving from ``EvaluationNodeConfig``
    and initialise node with it. Also ``EvaluationNode`` can be configured
    by ``JSONDict`` config. In this case no special config model should be
    created, but config should match ``EvaluationNodeConfig`` config model.

    Four most common evaluation operation are implemented, see methods
    with ``eval_`` prefix.
    """
    asyncable: bool = True
    config_model = EvaluationNodeConfig
    config: TEvaluationNodeConfig

    def __init__(self, config: Optional[TEvaluationNodeConfig] = None) -> None:
        """See the base class."""
        super().__init__(config=config)
        self._evaluations_map = defaultdict(list)
        conditions = self.config.conditions

        for var, evaluations in conditions.items():
            evaluations = evaluations if isinstance(evaluations, list) else [evaluations]

            for evaluation in evaluations:
                eval_dict = evaluation.dict()
                method_name = f'eval_{eval_dict.pop("eval_type")}'
                method = getattr(self, method_name)
                invert = eval_dict.pop('invert')
                handler = partial(method, **eval_dict)

                if invert:
                    def not_handler(value):
                        return not handler(value)
                    self._evaluations_map[var].append(not_handler)
                else:
                    self._evaluations_map[var].append(handler)

    def execute(self, state: State) -> NodeStatus:
        """Evaluates state variables.

        Returns SUCCESS status if all evaluations were successfully passed for all
        variables. Otherwise it returns:
        - FAILED status if at least one of the evaluations failed or at least one
            variable was not initialised and ``eval_none`` config flag is ``True``;
        - RUNNING status if ``eval_none`` config flag is ``False``, at least one variable
            being estimated is not initialised and ``None`` evaluations failed.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        result = NodeStatus.SUCCESS

        for var, evaluations in self._evaluations_map.items():
            value = state.vars.get(var)

            if value is None:
                if self.config.eval_none:
                    return NodeStatus.FAILED
                else:
                    result = NodeStatus.RUNNING
            else:
                for eval_handler in evaluations:
                    if not eval_handler(value):
                        return NodeStatus.FAILED

        return result

    @staticmethod
    def eval_equal(var: _EvalArgs, value: _EvalArgs) -> bool:
        """Evaluates if variable value is equal to the given value.

        Args:
            var: Value of the variable being evaluated.
            value: Value to compare.
        Returns:
            ``True`` if variable value is equal to the given value, else ``False``.
        """
        return var == value

    @staticmethod
    def eval_more(var: Union[int, float], value: Union[int, float], strict: bool = True) -> bool:
        """Evaluates if variable value is more then the given value.

        Args:
            var: Value of the variable being evaluated.
            value: Value to compare.
            strict: Strict comparison flag, if ``True`` - strict comparison is performed.
        Returns:
            ``True`` if variable value is more than the given value, else ``False``.
        """
        return var > value if strict else var >= value

    @staticmethod
    def eval_less(var: Union[int, float], value: Union[int, float], strict: bool = True) -> bool:
        """Evaluates if variable value is less then the given value.

        Args:
            var: Value of the variable being evaluated.
            value: Value to compare.
            strict: Strict comparison flag, if ``True`` - strict comparison is performed.
        Returns:
            ``True`` if variable value is less than the given value, else ``False``.
        """
        return var < value if strict else var <= value

    @staticmethod
    def eval_intersection(var: List[JSONSimpleType], value: List[JSONSimpleType]) -> bool:
        """Evaluates if list from variable value has at least one element in common with the given list.

        Elements order does not affect evaluation result, as it was checked two
        unordered sets intersection.

        Args:
            var: List variable value.
            value: List to search common elements with.
        Returns:
            ``True`` if list from variable value has at least one element in common
            with the given list, else ``False``.
        """
        return bool(set(var).intersection(set(value)))


class EvalNoneNodeConfig(ConfigModelBase):
    """Base model for the ``EvalNoneNode`` settings.

    Attributes:
        variables: List of str names of the variables to be evaluated to be initialised.
        invert: Boolean inversion flag. If ``True``, ``not`` operator is applied to the
            each variable evaluation result
    """
    variables: List[str] = Field(min_items=1)
    invert: bool = False


class EvalNoneNode(NodeBase):
    """Node that evaluates if state variables initialised.

    This node implements one of the most common tasks: check if some
    state variables values initialised. Please note that there is an
    important project-wide convention: ``None`` state variable value
    always represents uninitialised variable.
    """
    asyncable: bool = True
    config_model = EvalNoneNodeConfig
    config: EvalNoneNodeConfig

    def __init__(self, config: Optional[EvalNoneNodeConfig] = None) -> None:
        super().__init__(config=config)

    def execute(self, state: State) -> NodeStatus:
        """Evaluates if state variables initialised.

        If ``invert`` config flag is ``False`` returns FAILED status if at least one of the
        evaluated variables is uninitialised or is ``None``. Otherwise returns SUCCESS status.

        If ``invert`` config flag is ``True`` returns FAILED status if at least one of the
        evaluated variables is initialised and not ``None``. Otherwise returns SUCCESS status.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        if self.config.invert:
            for var in self.config.variables:
                if state.vars.get(var) is None:
                    return NodeStatus.FAILED
        else:
            for var in self.config.variables:
                if state.vars.get(var) is not None:
                    return NodeStatus.FAILED

        return NodeStatus.SUCCESS


class AssignNodeConfig(ConfigModelBase):
    """Base model for the ``SetNode`` settings.

    Every ``AssignNode`` config should be derived from this class. Each assignment
    of a value to a state variable is configured by an item of the ``assignments``
    field dict, where string key is equal to the variable name and value is
    assigned value.

    Attributes:
        assignments: Dict with variable naes and assigned values.
    """
    assignments: Dict[str, JSONType]


TAssignNodeConfig = TypeVar('TAssignNodeConfig', bound=AssignNodeConfig)


class AssignNode(NodeBase):
    """Node that assigns values to the state variables.

    ``AssignNode`` allows to parametrise all assignments of values to state variables,
    so its enough to create new config by deriving from ``AssignNodeConfig`` and
    initialise node with it. Also ``EvaluationNode`` can be configured by ``JSONDict``
    config. In this case no special config model should be created, but config should
    match ``AssignNodeConfig`` config model.
    """
    asyncable: bool = True
    config_model = AssignNodeConfig
    config: TAssignNodeConfig

    def __init__(self, config: Optional[TAssignNodeConfig] = None) -> None:
        """See the base class."""
        super().__init__(config=config)

    def execute(self, state: State) -> NodeStatus:
        """Assigns values to the state variables according to the config settings.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        for var, value in self.config.assignments.items():
            state.vars[var] = value

        return NodeStatus.SUCCESS


class ResetNodeConfig(ConfigModelBase):
    """Base model for the ``ResetNode`` settings.

    Attributes:
        variables: ``None`` or list of str names of the variables to be reset.
            If empty list or ``None``, ALL variables will be rest.
    """
    variables: Optional[List[str]] = Field(default=None)


class ResetNode(NodeBase):
    """Node that resets state valuables."""
    asyncable: bool = True
    config_model = ResetNodeConfig
    config: ResetNodeConfig

    def __init__(self, config: Optional[EvalNoneNodeConfig] = None) -> None:
        super().__init__(config=config)

    def execute(self, state: State) -> NodeStatus:
        """Resets state variables according to the config settings.

        Deletes items with keys given in config from the state variables dict.
        If variables names list in config is empty or ``None``, ALL variables will
        be deleted.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """

        if self.config.variables:
            for var in self.config.variables:
                state.vars.pop(var, None)
        else:
            state.vars.clear()

        return NodeStatus.SUCCESS



"""Base classes for the asynchronous implementations of the main node types.

``nodzz`` allows to implement asynchronous behavior trees and provides asynchronous
API. Here are the basic rules and principles of ``nodzz`` asynchronous development:
    - Each asynchronous tree node should have an asynchronous ``execute`` method
        implementation (i.e., each node ``execute`` method should be coroutine,
        see this for further references: https://docs.python.org/3/library/asyncio-task.html#awaitables)
        or it should be asyncable.
    - Asynchronous controller nodes should be derived from ``AsyncControllerNodeBase``
        class (or have this class in their ancestors chain) and must have an
        asynchronous ``execute`` method implementation.
    - Asynchronous task nods can ether have an asynchronous ``execute`` method
        implementation or be asyncable. Please see ``NodeBase`` class documentation
        for the further asyncable nodes reference.
    - ``AsyncTree`` should be used as an asynchronous behavior tree entry point.
    - All main ``nodzz`` node types are planned to have both synchronous and
        asynchronous (or asyncable) implementations.
    - Usually asynchronous behavior trees are used for the microservice pattern
        implementation: tree itself acts as an orchestrator and its nodes implement
        only IO-bound tasks while heavy CPU-bound tasks are implemented in the
        microservices which are called by the corresponding nodes.
"""
# TODO: Some of the docstrings parts should be moved to docs.

from inspect import iscoroutinefunction
from typing import Optional

from nodzz.basic_types import NodeStatus
from nodzz.core import State
from nodzz.nodes.base import NodeWrapperBase, ControllerNodeBase, ControllerConfig, TControllerConfig, TNode


class AsyncNodeWrapper(NodeWrapperBase):
    """Asynchronous node wrapper implementation.

    See the base class for the reference.
    """
    def __init__(self, node: TNode) -> None:
        """Initialises node wrapper.

        Mirrors wrapped node ``execute`` method or wraps it up in the debug
        executor. Also handles async wrapping of asyncable nodes ``execute``
        methods.

        Args:
            node: Initialised node instance to be wrapped.
        """
        super().__init__(node=node)

        if not iscoroutinefunction(node.execute):
            if node.asyncable:
                if self._tree_debug:
                    self.execute = self._execute_sync_debug
                else:
                    self.execute = self._execute_sync
            else:
                raise AttributeError(f'{node.__class__.__name__} can not be used in asynchronous trees')

        elif self._tree_debug:
            self.execute = self._execute_sync_debug

    async def execute(self, state: State) -> NodeStatus:
        """Calls wrapped node ``execute`` method."""
        return await self._node.execute(state=state)

    async def _execute_sync(self, state: State) -> NodeStatus:
        """Synchronously calls wrapped node ``execute`` method."""
        return self._node.execute(state=state)

    async def _execute_async_debug(self, state: State) -> NodeStatus:
        """Executes node and calls logging of its execution status.

        Tree debug mode wrapper for the node ``execute`` method.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = await self._node.execute(state=state)
        self._log_debug(status=status, state=state)
        return status

    async def _execute_sync_debug(self, state: State) -> NodeStatus:
        """Version of the ``_execute_async_debug`` with the synchronous node execution call."""
        status = self._node.execute(state=state)
        self._log_debug(status=status, state=state)
        return status


class AsyncControllerNodeBase(ControllerNodeBase):
    """Base class for asynchronous controller nodes.

    See the base class for the reference.
    """
    config_model = ControllerConfig

    def __init__(self, *children: TNode, config: Optional[TControllerConfig] = None) -> None:
        """See the base class."""
        super().__init__(config=config)
        self.children = tuple(AsyncNodeWrapper(child) for child in children)

    async def execute(self, state: State) -> NodeStatus:
        """See the base class."""
        raise NotImplementedError



"""Asynchronous controller nodes implementations."""

from nodzz.core import State
from nodzz.basic_types import NodeStatus
from nodzz.async_nodes.base import AsyncControllerNodeBase, TNode


class AsyncSelectorNode(AsyncControllerNodeBase):
    """Asynchronous implementation of the non-persistent selector node.

    See ``SelectorNode`` class for the reference.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    async def execute(self, state: State) -> NodeStatus:
        """See ``SelectorNode.execute`` method for the reference."""
        status = NodeStatus.FAILED

        for node in self.children:
            status = await node.execute(state)

            if status != NodeStatus.FAILED:
                break

        return status


class AsyncPersistentSelectorNode(AsyncControllerNodeBase):
    """Asynchronous implementation of the persistent selector node.

    See ``PersistentSelectorNode`` class for the reference.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    async def execute(self, state: State) -> NodeStatus:
        """See ``PersistentSelectorNode.execute`` method for the reference."""
        status = NodeStatus.FAILED

        for node in self.children:
            current_status = state.get_node_status(node_id=node.id)

            if current_status == NodeStatus.READY or current_status == NodeStatus.RUNNING:
                status = await node.execute(state)

            if status == NodeStatus.FAILED or status == NodeStatus.RUNNING:
                state.set_node_status(node_id=node.id, status=status)

            if status == NodeStatus.RUNNING or status == NodeStatus.SUCCESS:
                break

        if status == NodeStatus.SUCCESS or status == NodeStatus.FAILED:
            self.reset(state)

        return status


class AsyncSequenceNode(AsyncControllerNodeBase):
    """Asynchronous implementation of the non-persistent sequence node.

    See ``SequenceNode`` class for the reference.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    async def execute(self, state: State) -> NodeStatus:
        """See ``SequenceNode.execute`` method for the reference."""
        status = NodeStatus.SUCCESS

        for node in self.children:
            status = await node.execute(state)

            if status != NodeStatus.SUCCESS:
                break

        return status


class AsyncPersistentSequenceNode(AsyncControllerNodeBase):
    """Asynchronous implementation of the persistent sequence node.

    See ``PersistentSequenceNode`` class for the reference.
    """
    def __init__(self, *children: TNode, config=None) -> None:
        """See the base class."""
        super().__init__(*children, config=config)

    async def execute(self, state: State) -> NodeStatus:
        """See ``PersistentSequenceNode.execute`` method for the reference."""
        status = NodeStatus.SUCCESS

        for node in self.children:
            current_status = state.get_node_status(node_id=node.id)

            if current_status == NodeStatus.READY or current_status == NodeStatus.RUNNING:
                status = await node.execute(state)

            if status == NodeStatus.SUCCESS or status == NodeStatus.RUNNING:
                state.set_node_status(node_id=node.id, status=status)

            if status == NodeStatus.RUNNING or status == NodeStatus.FAILED:
                break

        if status == NodeStatus.SUCCESS or status == NodeStatus.FAILED:
            self.reset(state)

        return status



"""Behavior tree management instruments."""

from typing import Optional, Dict, Union, Type

from nodzz.async_nodes.base import AsyncNodeWrapper
from nodzz.basic_types import NodeStatus
from nodzz.config import ConfigSet
from nodzz.core import State
from nodzz.nodes.base import TNode, NodeWrapperBase, NodeWrapper, NodeBase, ControllerNodeBase
from nodzz.utils import import_by_name


def _make_tree(config_set: ConfigSet, node_name: str, nodes: Optional[Dict[str, TNode]] = None) -> NodeBase:
    """Makes behavior tree.

    Recursively initialises nodes and composes behavior tree from them.

    Args:
        config_set: ConfigSet instance with tree nodes config.
        node_name: Str name of the behavior tree root node.
        nodes: Dict with initialised nodes instances. Keys are nodes names,
            values are nodes instances. Used for accessing already initialised
            nodes throughout function recursive calls to prevent repeated
            initialisations of the same node. Should be ``None`` when the
            method is called not from itself.

    Returns:
        Initialised behavior tree inside its root node. Root node returned
        without node wrapper.
    """
    nodes = nodes or {}
    result = nodes.get(node_name)

    if not result:
        config = config_set.get_config(node_name)
        class_name = config['class_name']
        node_class = import_by_name(class_name)

        if issubclass(node_class, ControllerNodeBase):
            children = [_make_tree(config_set, node_name, nodes) for node_name in config['children']]
            result = node_class(*children, config=config)
        elif issubclass(node_class, NodeWrapperBase):
            raise TypeError(f'Node wrapper can not be used as node class')
        elif issubclass(node_class, NodeBase):
            result = node_class(config=config)
        else:
            raise TypeError(f'Wrong component type: {class_name}, only NodeBase subclasses allowed')

        nodes[node_name] = result

    return result


class _TreeBase:
    """Base class for the behavior tree entry point.

    Tree handles behavior tree (re)initialisation and provides entry point
    for its execution.
    """
    def __init__(self, config_set: ConfigSet, wrapper_class: Type[Union[NodeWrapper, AsyncNodeWrapper]]) -> None:
        """Initialises behavior tree entry point.

        Args:
            config_set: ConfigSet with behavior tree nodes configs.
        """
        self._config_set = config_set
        self._config_set.resolve_configs()
        self._wrapper_class = wrapper_class
        self._root_node: Optional[Union[NodeWrapper, AsyncNodeWrapper]] = None

    def init_tree(self, node_name: str) -> None:
        """Initialises behavior tree.

        Deinitialises previously initialised tree if such exists.

        Args:
            node_name: Str name of the behavior tree root node.
        """
        del self._root_node
        root_node = _make_tree(config_set=self._config_set, node_name=node_name)
        self._root_node = self._wrapper_class(root_node)
        self._root_node.prepare(node_id='0')

    def execute(self, state: State) -> NodeStatus:
        """Executes behavior tree root node.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node statuses: SUCCESS, FAILED, RUNNING.
        """
        raise NotImplementedError


class Tree(_TreeBase):
    """Synchronous behavior tree entry point."""
    def __init__(self, config_set: ConfigSet) -> None:
        """See the base class."""
        super().__init__(config_set=config_set, wrapper_class=NodeWrapper)

    def execute(self, state: State) -> NodeStatus:
        """See the base class."""
        return self._root_node.execute(state)


class AsyncTree(_TreeBase):
    """Asynchronous behavior tree entry point."""
    def __init__(self, config_set: ConfigSet) -> None:
        """See the base class."""
        super().__init__(config_set=config_set, wrapper_class=AsyncNodeWrapper)

    async def execute(self, state: State) -> NodeStatus:
        """See the base class."""
        return await self._root_node.execute(state)



"""Base classes for the main node types."""
# TODO: Some of the docstrings parts should be moved to docs.

from os import getenv
from typing import TypeVar, Optional, List

from nodzz.basic_types import NodeStatus
from nodzz.core import ConfigModelBase, ComponentBase, State, TConfig

_tree_debug_env = getenv('NODZZ_TREE_DEBUG')
_tree_debug = True if _tree_debug_env == '1' else False


class NodeBase(ComponentBase):
    """Base class for the node entity.

    Node is the main (and only) building block of behavior trees. In most
    cases each node performs some of these tasks when executed:
    - Control execution of other nodes (controller nodes);
    - Get information from external environment;
    - Update or modify state variables;
    - Estimate state variables;
    - Affect external environment.

    All nodes implementations or other nodes abstractions (except controller
    nodes) should derive from this class.

    Attributes:
        asyncable: Bool `asyncio` compatibility flag. If ``True`` and node ``execute``
            method is not defined as ``async``, it will be automatically wrapped to
            the coroutine during asynchronous behavior tree initialisation. Otherwise
            exception will be risen if synchronous node will tried to be used in
            asynchronous tree. Such behavior is implemented to allow synchronous nodes
            be reused in asynchronous trees. Please be cautious with this flag and set
            it to ``True`` only when implementing nodes with non CPU-bound ``execute``
            method.
    """
    asyncable: bool = False

    def __init__(self, config: Optional[TConfig] = None) -> None:
        """See the base class."""
        super().__init__(config=config)

    def execute(self, state: State) -> NodeStatus:
        """Executes node.

        When implemented, method executes node logic and returns its execution
        status:
        - SUCCESS status means that the node successfully performed its task;
        - FAILED status means that the node was not able to perform its task
            successfully. Either because of behavior agent internal or external
            environment state or because of application internal error.
        - RUNNING status that the node currently performing some actions (or
            awaiting something) to return SUCCESS or FAILED status further.
            RUNNING status interpretation depends of the internal logic of the
            entity which controls execution of the node which returned this
            status.

        The fact, that the final node execution status is either SUCCESS or FAILED,
        allows to consider any node as a logical gate that affects decision tree
        execution.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        raise NotImplementedError

    def prepare(self, node_id: str) -> None:
        """Method is implemented in ``NodeWrapper`` and ``ControllerNodeBase`` classes
        and resides here for API compatibility.
        """
        pass

    def reset(self, state: State) -> None:
        """Method is implemented in ``NodeWrapper`` and ``ControllerNodeBase`` classes
        and resides here for API compatibility.
        """
        pass


TNode = TypeVar('TNode', bound=NodeBase)


class NodeWrapperBase(NodeBase):
    """Node wrapping container.

    Each initialised node is wrapped to this container before it is allocated
    to its position in the behavior tree. The may reason for this is that
    each node is identified by its position in the tree and sometimes the
    same initialised behavior should be executed from the different nodes of
    the tree.

    So when we put one initialised node (functional instance) in several
    places (logical nodes) in the tree, several NodeWrapper instances are
    created (each instance for an each logical node). Each wrapper executes
    the same functional instance but handles execution statuses according
    its position in the tree.

    NodeWrapper mirrors standard Node API.
    """
    def __init__(self, node: TNode) -> None:
        """Initialises node wrapper.

        Args:
            node: Initialised node instance to be wrapped.
        """
        super().__init__(config=node.config)
        self.id: Optional[str] = None
        self._node = node
        self._tree_debug = _tree_debug

    def execute(self, state: State) -> NodeStatus:
        """Calls wrapped node ``execute`` method."""
        raise NotImplementedError

    # TODO: Refactor, use logging.
    def _log_debug(self, status: NodeStatus, state: State) -> None:
        """Logs node execution status and behavior tree state after node execution.

        Args:
            status: Node execution status.
            state: Behavior tree execution state.
        """
        state_str = str(state.to_dict())
        config_dict = dict(self.config)
        name = config_dict.get('name')
        debug_str = f'[nodzz debug] id: {self.id}, name: {name}, status: {str(status)}, state: {state_str}'
        print(debug_str)

    def prepare(self, node_id: str) -> None:
        """Sets node id.

        Also calls wrapped node ``prepare`` method. If this is controller node,
        all child nodes ``prepare`` methods will be subsequently called.

        Args:
            node_id: String node id.
        """
        self.id = node_id
        self._node.prepare(node_id)

    def reset(self, state: State) -> None:
        """Resets node status.

        Sets node status to READY. Also calls wrapped node ``reset`` method.
        If this is controller node, all child nodes statuses will be
        subsequently reset.

        Args:
            state: Behavior tree execution state.
        """
        state.reset_node_status(node_id=self.id)
        self._node.reset(state)


class NodeWrapper(NodeWrapperBase):
    """Synchronous node wrapper implementation.

    For more reference see the base class.
    """
    def __init__(self, node: TNode) -> None:
        """Initialises node wrapper.

        Mirrors wrapped node ``execute`` method or wraps it up in the debug
        executor.

        Args:
            node: Initialised node instance to be wrapped.
        """
        super().__init__(node=node)

        if self._tree_debug:
            self.execute = self._execute_debug

    def execute(self, state: State) -> NodeStatus:
        """Calls wrapped node ``execute`` method."""
        return self._node.execute(state=state)

    def _execute_debug(self, state: State) -> NodeStatus:
        """Executes node and calls logging of its execution status.

        Tree debug mode wrapper for the node ``execute`` method.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        status = self._node.execute(state=state)
        self._log_debug(status=status, state=state)
        return status


class ControllerConfig(ConfigModelBase):
    children: List[str] = []


TControllerConfig = TypeVar('TControllerConfig', bound=ControllerConfig)


class ControllerNodeBase(NodeBase):
    """Base class for control flow (controller) nodes.

    Controller nodes serve as a containers for another nodes (children nodes)
    and have one main purpose: control execution of its child nodes. Each time
    control node is executed, it either decides to execute one or several of its
    child nodes or to return its own execution status to its parent (another
    control node). In fact, this way control nodes implement branching in
    behavior trees. There are two base concepts, which form control nodes
    execution mechanism: Child nodes execution order and child nodes execution
    statuses.

    Each controller node is initialised with an ORDERED set of its children.
    That means that every initialised controller is always given the default order
    of its child nodes execution, though some controller node implementations
    can deliberately ignore this order.

    This concept maps on rational beings (both natural and artificial) behavior
    pattern: every rational being tends to prioritize its goals (and, as a result,
    behaviors aimed to achieve them). And every rational being tends to prioritize
    actions aimed to achieve any of its goals.

    Also control node does not supposed to have access to behavior state and makes
    decision about its next action based only on its already executed child nodes
    statuses. That means that agent behaviors stay fully encapsulated in nodes
    (non-controller) while decision making process is moved to the tree structure.
    This allows to easily add or remove nodes to the tree and simplifies subtrees
    reusing.

    All control nodes implementations or other control nodes abstractions should
    derive from this class.

    Attributes:
        children: Tuple of node wrappers with initialised child nodes.
    """
    config_model = ControllerConfig

    # TODO: Think of unifying __init__ signature among all components.
    def __init__(self, *children: TNode, config: Optional[TControllerConfig] = None) -> None:
        """Initialises controller node.

        Args:
            *children: Initialised child nodes.
            config: Pydantic BaseSettings node config.
        """
        super().__init__(config=config)
        self.children = tuple(NodeWrapper(child) for child in children)

    def execute(self, state: State) -> NodeStatus:
        """Method should implement execution logic of control node child nodes.

        Attention! Child nodes statuses management should be also implemented here.

        Args:
            state: Behavior tree execution state.

        Returns:
            One of three node execution statuses: SUCCESS, FAILED, RUNNING.
        """
        raise NotImplementedError

    def prepare(self, node_id: str) -> None:
        """Assigns ids to the children nodes.

        Args:
            node_id: String node id.
        """
        for i, child in enumerate(self.children):
            child_id = f'{node_id}.{i}'
            child.prepare(node_id=child_id)

    def reset(self, state: State) -> None:
        """Resets children nodes states.

        Args:
            state: Behavior tree execution state.
        """
        for child in self.children:
            child.reset(state)



"""Simple utils and boilerplate code."""

import json
from importlib import import_module
from pathlib import Path
from typing import Any

import yaml

from nodzz.basic_types import JSONType


def load_yaml(file_path: Path) -> JSONType:
    """Reads YAML file contents into python object.

    Args:
         file_path: Yaml file path in pathlib format.

    Returns:
         Yaml file contents loaded into python object.
    """
    with file_path.open('r', encoding='utf8') as f:
        result = yaml.safe_load(f)

    return result


def load_json(file_path: Path) -> JSONType:
    """Reads JSON file contents into python object.

        Args:
             file_path: Json file path in pathlib format.

        Returns:
             Json file contents loaded into python object.
        """
    with file_path.open('r', encoding='utf8') as f:
        result = json.load(f)

    return result


def load_file(file_path: Path) -> JSONType:
    """Reads JSON and YAML contents files into Python object.

    Method wraps JSON and YAML files load.

    Args:
        file_path: ``pathlib`` object with configuration file path.

    Returns:
        Object with loaded file contents.
    """
    if file_path.suffix == '.json':
        return load_json(file_path)
    elif file_path.suffix in ['.yaml', '.yml']:
        return load_yaml(file_path)
    else:
        raise ValueError(f'Unsupported extension of file: {str(file_path)}, can handle only .yaml, .yml or .json')


def import_by_name(name: str) -> Any:
    """Imports Python object using its full path in str format.

    Args:
        name: Object full string path, compiled of dot-delimited modules names
            and object name.

    Returns:
        Imported object.
    """
    module_name = name.rsplit('.', 1)
    object_name = module_name.pop(-1)
    module_name = module_name[0] if module_name else '__init__'
    return getattr(import_module(module_name), object_name)
